/**
 * What this sketch does
 * 
 * - Reads current time from DS3231
 * - Reads umidity from DHT11/DHT22 Humidity/Temperature sensor
 * - Reads light information photo-diode with the help of the emission from a LED (FIXME: Specs?)
 * - Reads temperature from DS18B20
 * 
 * Libraries required - Unless otherwise noted, they are available from Arduino IDE (Sketch -> Include Libraries -> Manage Libraries)
 * 
 * DS3231 1.0.0 for DS3231
 * SimpleDHT 1.0.6 for DHT
 * DallasTemperature 3.7.6 for DS18B20
 * OneWire 2.3.3 for DS18B20
 * Wire (part of the Arduino IDE, no installation required)
 * 
 */

#include "constants.h"
#include "globals.h"
#include "utils.h"
#include "commands.h"



/*******************
 * Humidity Sensor *
 *******************/
// Note 1: We use SimpleDHT instead of Adafruit's library; the latter doesn't work, can't find the device
// Note 2: We only use humidity from this sensor, precision too low for temperature
#include <SimpleDHT.h>
SimpleDHT22 dht;

/******************************
 * Digital Temperature Sensor *
 ******************************/
#include <OneWire.h>
#include <DallasTemperature.h>
OneWire oneWire(TEMPERATURE_PIN);
DallasTemperature sensors(&oneWire);

#include <Wire.h>





void setup() {

  // SERIAL PORT SETUP
  // initialize serial communication at 9600 bits per second:
  Serial.begin(9600);

  // LED SETUP
  // initialize the digital pins as an output.
  pinMode(LED_PD1_PIN, OUTPUT);
  pinMode(LED_PD2_PIN, OUTPUT);
  pinMode(PD1_POWER_PIN, OUTPUT);
  pinMode(PD2_POWER_PIN, OUTPUT);
  digitalWrite(PD1_POWER_PIN, LOW);
  digitalWrite(PD2_POWER_PIN, LOW);
  
  // set the internal reference voltage 1.1V
  analogReference(INTERNAL);

  // HUMIDITY SENSOR SETUP
  // dht.begin();
  Wire.begin();
  measure=0;
  
  // TEMPERATURE SENSOR SETUP
  sensors.begin();

}

void printHumidity() {
  float temperature = 0;
  float humidity = 0;
  Serial.print(",Hum,");
  int err = SimpleDHTErrSuccess;
  if ((err = dht.read2( DHT_PIN , &temperature, &humidity, NULL)) != SimpleDHTErrSuccess) {
    Serial.print("ERR "); 
    Serial.print(err);
    Serial.print(",");
    delay(1000);
  } else {
    Serial.print(humidity);
    Serial.print(",%");
  }
}

void printTemperature() {
    sensors.requestTemperatures();

    Serial.print(",Temp,");
    int8_t temperature = sensors.getTempCByIndex(0);
    if (temperature == -127) {
      Serial.print("ERR,C");
    } else {    
      Serial.print(sensors.getTempCByIndex(0));
      Serial.print(",C");
    }
}

void waitStart() {
  debug("Waiting for start command");
  while(!started) {
    waitCharacter();
    handleCommands();
  }
}

void printCounter(int measure) {
  // Iteration counter
  Serial.print("N:,");
  Serial.print(measure);
  Serial.print(",");
}



/*****************************************************************************
 * Read light information from two photo diodes with light from a single led *
 *****************************************************************************/
void readLightSensorTwoPd(int ledPin, int pd1Pin, int pd2Pin, int pd1PowerPin, int pd2PowerPin, double* pd1SensorValue, double* pd2SensorValue, int ledValue, int pdOffValue) {
  int ledOffValue = LOW;
  if (LED_ALWAYS_ON) {
    ledValue = HIGH;
    ledOffValue = HIGH;
  }
  // Turn on the led (if ledPin is not an off value) and the two photo diodes. Wait a few seconds for stabilization
  digitalWrite(ledPin, ledValue);   
  digitalWrite(pd1PowerPin, HIGH);
  digitalWrite(pd2PowerPin, HIGH);
  delay(3000);

  // Read the values from the light sensor. Take the mean value out of a number of measures with 1 ms delay between them
  *pd1SensorValue = 0;
  *pd2SensorValue = 0;
  for (int x=0; x< LIGHT_SENSOR_N_READS ; x++) {
    *pd1SensorValue = *pd1SensorValue + analogRead(pd1Pin);
    *pd2SensorValue = *pd2SensorValue + analogRead(pd2Pin);
    delay(1);         // delay for stability
  }

  // turn off the led. Potentially turn off the two PD (pdOffValue may in fact be HIGH if we need to keep them on)
  digitalWrite(ledPin, ledOffValue);   
  digitalWrite(pd1PowerPin, pdOffValue);
  digitalWrite(pd2PowerPin, pdOffValue);
  
  *pd1SensorValue = (*pd1SensorValue)*MV_PER_STEP / LIGHT_SENSOR_N_READS;
  *pd2SensorValue = (*pd2SensorValue)*MV_PER_STEP / LIGHT_SENSOR_N_READS;
}




/*****************************************************************************
 * Read light information from a single photo diode with light from a single led *
 *****************************************************************************/
double readLightSensor(int ledPin, int pdPin, int pdPowerPin, int ledValue, int pdOffValue) {
  int ledOffValue = LOW;
  if (LED_ALWAYS_ON) {
    ledValue = HIGH;
    ledOffValue = HIGH;
  }

  double sensorValue;
  // Turn on the led, then wait 3 seconds for stabilization
  digitalWrite(ledPin, ledValue);   
  digitalWrite(pdPowerPin, HIGH);
  delay(3000);

  // Read the values from the light sensor. Take the mean value out of a number of measures with 1 ms delay between them
  sensorValue = 0;
  for (int x=0; x< LIGHT_SENSOR_N_READS ; x++) {
    sensorValue = sensorValue + analogRead(pdPin);
    delay(1);         // delay for stability
  }

  digitalWrite(ledPin, ledOffValue);   
  digitalWrite(pdPowerPin, pdOffValue);
  
  sensorValue = sensorValue*MV_PER_STEP / LIGHT_SENSOR_N_READS;
  return sensorValue;
}



/*
 * Reading with double led. To make the code short the loop is called two times. The first (secondPart == false)
 * we read from PD1 (index 0), the second from the PD2 (index 1)
 */
void loop2Led() {
  double baseLine;
  double sensorValue;
  int i = secondPart ? 1 : 0;
  
  // Light Sensor
  // Baseline: keep off led, turn leave on pd after reading
  
  baseLine = readLightSensor(ledPins[i], pdPins[i], pdPowerPins[i], LOW, HIGH);

  // sensor value: turn on led, keep on pd after reading
  sensorValue = readLightSensor(ledPins[i], pdPins[i], pdPowerPins[i], HIGH, HIGH);
  
  // print out the values we just read:
  Serial.print(",");
  Serial.print(diodeDescription[i]);
  Serial.print(" (baseline),");
  Serial.print(baseLine);
  Serial.print(",mV");

  Serial.print(",");
  Serial.print(diodeDescription[i]);
  Serial.print(",");
  Serial.print(sensorValue);
  Serial.print(",mV");

  // final read: turn off led, turn off pd after reading
  baseLine = readLightSensor(ledPins[i], pdPins[i], pdPowerPins[i], LOW, LOW);
  Serial.print(",");
  Serial.print(diodeDescription[i]);
  Serial.print(" (baseline),");
  Serial.print(baseLine);
  Serial.print(",mV");
  secondPart = !secondPart;
}












/*
 * Readings with a single led. What we do is to read in a single pass from both photo diodes
 */
void loop1Led() {
  double pd1SensorValue;
  double pd1BaseLine;
  double pd1BaseLine2;
  double pd2SensorValue;
  double pd2BaseLine;
  double pd2BaseLine2;
  // Light Sensor
  readLightSensorTwoPd(ledPins[0], pdPins[0], pdPins[1], pdPowerPins[0], pdPowerPins[1], &pd1BaseLine, &pd2BaseLine, LOW, HIGH);
  readLightSensorTwoPd(ledPins[0], pdPins[0], pdPins[1], pdPowerPins[0], pdPowerPins[1], &pd1SensorValue, &pd2SensorValue, HIGH, HIGH);
  readLightSensorTwoPd(ledPins[0], pdPins[0], pdPins[1], pdPowerPins[0], pdPowerPins[1], &pd1BaseLine2, &pd2BaseLine2, LOW, LOW);
  // print out the values we just read:
  Serial.print(",");
  Serial.print(diodeDescription[0]);
  Serial.print(" (baseline),");
  Serial.print(pd1BaseLine);
  Serial.print(",mV");

  Serial.print(",");
  Serial.print(diodeDescription[0]);
  Serial.print(",");
  Serial.print(pd1SensorValue);
  Serial.print(",mV");

  Serial.print(",");
  Serial.print(diodeDescription[0]);
  Serial.print(" (baseline),");
  Serial.print(pd1BaseLine2);
  Serial.print(",mV");


  
  Serial.print(",");
  Serial.print(diodeDescription[1]);
  Serial.print(" (baseline),");
  Serial.print(pd2BaseLine);
  Serial.print(",mV");

  Serial.print(",");
  Serial.print(diodeDescription[1]);
  Serial.print(",");
  Serial.print(pd2SensorValue);
  Serial.print(",mV");

  Serial.print(",");
  Serial.print(diodeDescription[1]);
  Serial.print(" (baseline),");
  Serial.print(pd2BaseLine2);
  Serial.print(",mV");
  
}

/******
 * This is only for test mode
 */
#define TEST_ARRAY_SIZE 100
// 0 : before turning on led
// 1 : with led on
// 2 : after turning off led
int testLoopPart = 0;
int testLoopN = 0;
float reads[TEST_ARRAY_SIZE];

/*
 * Like readLightSensor, but without warmup
 */
double readLightSensorTest(int pdPin) {
  int nReads = 10;
  double sensorValue;
  // Read the values from the light sensor. Take the mean value out of a number of measures with 1 ms delay between them
  sensorValue = 0;
  for (int x=0; x< nReads ; x++) {
    sensorValue = sensorValue + analogRead(pdPin);
    delay(1);         // delay for stability
  }
  sensorValue = sensorValue*MV_PER_STEP / nReads;
  return sensorValue;
}



void testLoop() {

  if (testLoopN == 0 ) {
    // new round, init the value
    if (testLoopPart == 0) {
      // Ensure the led is off, wait some time just in case
      digitalWrite(PD1_POWER_PIN, LOW);
      delay(3000);
      digitalWrite(LED_PD1_PIN, HIGH);
    } else if (testLoopPart == 1) {
      // Ensure led is on
      digitalWrite(PD1_POWER_PIN, HIGH);
      //delay(3000);
    } else if (testLoopPart == 2) {
      digitalWrite(PD1_POWER_PIN, LOW);
      // delay(3000);
    }
  }
  
  double sensorValue = 0;
  sensorValue = readLightSensorTest(PD1_PIN);
  reads[testLoopN] = sensorValue;

  testLoopN += 1;
  if (testLoopN == TEST_ARRAY_SIZE) {
    // End of the tour, print all the data
    const DateTime& now = readDateTime();
    for (int i=0; i<TEST_ARRAY_SIZE; ++i) {
      Serial.print("N:,");
      Serial.print(measure+i);
      Serial.print(",");
      printDate(now);
      Serial.print(", Round,");
      Serial.print(testLoopPart);
      Serial.print(",,Stage,");
      switch(testLoopPart) {
        case 0:
          Serial.print( "before,," );
          break;
        case 1:
          Serial.print( "during,," );
          break;
        case 2:
          Serial.print("after,,");
          break;
        default:
          Serial.print("???,,");
      }
      Serial.print("Value,");
      Serial.print(reads[i]);      
      Serial.println(",mV");
    }
    Serial.println("");

    // check if we are at the end of the test, reset everythin in case
    testLoopN = 0;
    testLoopPart += 1;
    if (testLoopPart == 3) {
      delay(CALIBRATION_LOOP_WAIT);               // wait for the next measure
    }
  }
}

void loop() {
  if (!started && !secondPart) {
    waitStart();      
  }
  if (numberOfLeds ==0 ){
    testLoop();
    return;
  }
  if (!secondPart) {
    Serial.println("# Starting loop");
    printCounter(measure++);
  }
  
  // Current Date/Time
  const DateTime& now = readDateTime();
  printDate(now);

  if (numberOfLeds == 1 ) {
    loop1Led();
  } else if (numberOfLeds == 2 ) {
    loop2Led();
  }
  
  printHumidity();
  printTemperature();

  // Print the end of line only if we are using a single led or we have finished the second part of the two led case 
  // Note that the flag secondPart here is the opposite of the value at the beginning since we have changed it inside the loop
  if (numberOfLeds == 1 || !secondPart) {
    Serial.println("");
    handleCommands();
  }
  long int waitMs= waits[waitChoice];
  if (numberOfLeds == 2) {
    waitMs /= 2;
  }
  delay(waitMs);               // wait for the next measure
}

